=begin

Copyright 2011, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESSED OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         MakeProjectiveUnique.rb
Author:       Andreas Eisenbarth
Usage:        Select faces with the same texture,
              menu Plugins > 'Make Projective Textures Unique'
              faces selected => context menu > 'Make Projective Textures Unique'
Description:  This Plugin allows to make texture unique on many faces. This is useful 
              for exporting to fileformats that don't allow projectively distorted textures.
              Such a feature is not available in the API and this plugin makes use
              of SketchUp's texture writer which "bakes" projective distortions 
              into images, but leaves undistorted or affine textures unchanged.

Version:      1.0
Date:         21.11.2011

=end
require 'sketchup.rb'



module AE

module AE::MakeProjectiveUnique
@dir = File.dirname(__FILE__) # this plugin's folder
@temp = ( ENV['TMPDIR'] || ENV['TMP'] || ENV['TEMP'] ).gsub('\\','/')
@platform = RUBY_PLATFORM.include?('darwin')? "mac" : (RUBY_PLATFORM.include?('mswin')? (!ENV["WINEPREFIX"].nil?? "wine" : "mswin") : "other")

unless file_loaded?(File.basename(__FILE__))
  cmd_makeunique_single = UI::Command.new("Make Projective Textures Unique"){AE::MakeProjectiveUnique.each_single_face}
  cmd_makeunique_single.menu_text = "Make Projective Textures Unique"
  # Menu
	menu_plugins = UI.menu("Plugins")
	menu_plugins.add_item(cmd_makeunique_single)

	UI.add_context_menu_handler do |menu|
		if !Sketchup.active_model.selection.empty? && Sketchup.active_model.selection.find{|x| x.class==Sketchup::Face && (x.material&&x.material.materialType>0||x.back_material&&x.back_material.materialType>0)}
			menu.add_item(cmd_makeunique_single)
		end
  end
end



def self.each_single_face
  entities = Sketchup.active_model.selection.to_a
  return UI.messagebox("Select textured faces that you want to make unique.") if entities.empty?
  Sketchup.active_model.start_operation("Make each texture unique")
  entities.each{|e|
    next if e.class != Sketchup::Face
    self.single(e, true) if e.material
    self.single(e, false) if e.back_material
  }
  Sketchup.active_model.commit_operation
end #def each_single_face



##
# Replace the texture on one face with a new, unique texture
#
def self.single(face, front=true)
  return puts("AE::MakeProjectiveUnique: no face given") if face.nil? || face.class != Sketchup::Face
  return puts("AE::MakeProjectiveUnique: face has no texture") if (front && !face.material || !front && !face.back_material) || (front && face.material && face.material.materialType==0 || !front && face.back_material && face.back_material.materialType==0)
  @model = Sketchup.active_model
  @tw = Sketchup.create_texture_writer
  return if !self.texture_distorted?(face, front)
  begin
    m = (front)? face.material : face.back_material
    # export texture
    texturepath = self.export_texture(m, face, front)
    return if texturepath.nil?
    # get bounding box with 2 degrees of freedom
    b = AE::Bounds2.new(face)
    points = [b.corner(2), b.corner(0), b.corner(1), b.corner(3)] # counter-clockwise from top left
    # get UV coordinates
    uv_help = face.get_UVHelper(true, true, @tw)
    uvqs = points.collect{|p| (front)? uv_help.get_front_UVQ(p) : uv_help.get_back_UVQ(p)}
    uvqs.each{|uv| uv.x/=uv.z; uv.y/=uv.z; uv.z=1}
    # import new texture
    new_material = self.import_texture(texturepath, m)
    new_material.texture.size = [b.width, b.height]
    # position texture
    pt_array = [points[0], [0,1,1], points[1], [0,0,1], points[2], [1,0,1], points[3], [1,1,1]]
    face.position_material(new_material, pt_array, front)
    File.delete(texturepath) if File.exists?(texturepath)
  rescue Exception => e
    File.delete(texturepath) if texturepath && File.exists?(texturepath)
    @model.abort_operation
    puts("Error #<#{e.class.name}:#{e.message}.>")
    puts(e.backtrace) if $VERBOSE
  end #begin/rescue
end #def process



##
# export a texture to texturepath
#
def self.export_texture(material, face=nil, front=true)
  texturepath = self.to_filename(@temp, File.basename(material.texture.filename))
  # if no face given, create temporary group to export texture image without uv-mapping
  (tmp_group = @model.entities.add_group).material = material if face.nil?
  (face)? @tw.load(face, front) : @tw.load(tmp_group)
  success = (face)? @tw.write(face, front, texturepath) : @tw.write(tmp_group, texturepath)
  puts(material.name+": could not write file "+texturepath) if success!=0
  @model.entities.erase_entities(tmp_group) if face.nil?
  return (success==0)? texturepath : nil
end #def import_texture



##
# import a texture from texturepath, inherit properties from similar material
#
def self.import_texture(texturepath, similar_m=nil)
  return (puts("file "+texturepath+" not found"); similar_m) if !File.exists?(texturepath)
  i = 0; new_name = (similar_m)? similar_m.display_name : File.basename(texturepath, File.extname(texturepath))
  (new_name = similar_m.display_name + i.to_s; i+=1) while @model.materials[new_name]
  m = @model.materials.add(new_name)
  m.texture = texturepath
  if similar_m
    m.texture.size = [similar_m.texture.width, similar_m.texture.height]
    #m.color = similar_m.color
    m.alpha = similar_m.alpha
  end
  return m
end #def import_texture



## 
# test if a face has a projective (not affine) distortion
#   based on Tomasz Marek's su2kt.rb: SU2KT.texture_distorted
#   SketchUp's texturewriter writes a unique file for projective distortions.
#   Thus a projective texture gets a different handle, 
#   while two affine textures get the same handle.
#
def self.texture_distorted?(face, front=true)
  return false if face.class!=Sketchup::Face
	distorted=false
	temp_tw = Sketchup.create_texture_writer
  ents = @model.active_entities
	@model.start_operation "Group"
	group = ents.add_group
	group.material = (front==true)? face.material : face.back_material
	group_handle = temp_tw.load(group)
	temp_handle = temp_tw.load(face, front)
	ents.erase_entities(group)
	Sketchup.undo
	distorted = true if temp_handle != group_handle
	temp_tw = nil
	return distorted
end



##
# Convert a random string into a proper, unique filename
# arguments:
#   dir: directory where the file should go
#   string: any user input
#   ext: force a specific extension if the string doesn't contain one
# returns: a complete file path
# 
def self.to_filename(dir, string, ext=nil)
  # separate extension from basename
  ext = string[/\.\w{1,3}$/].to_s if ext.nil?
  string.sub!(/#{ext}$/,"")
  # clean up basename
  string = string[/.{1,30}/].gsub(/\n|\r/,"").gsub(/[^0-9a-zA-Z\-\_\.\(\)\#]+/,"_").gsub(/^_+|_+$/,"").to_s
  string = "file" if string.empty?
  base = File.join(dir, string)
  # detect collision of filenames and return alternative filename (numbered)
  base_orig = base
  i = 0
  while File.exists?(base + ext)
    base = base_orig + i.to_s
    i += 1
  end
  return base + ext
end #def to_filename



end #module AE::MakeProjectiveUnique



class AE::Bounds2

  ##
  # Fits a BoundingBox to a face with 2 degrees of freedom (heading and tilt, but roll=0)
  # This type of BoundingBox is what SketchUp uses for example for the orientation of "unique textures".
  #
  def initialize(*ents)
    return if ents.flatten.empty?
    ents.flatten! if ents[0].class == Array
    face = ents.find{|e| e.class==Sketchup::Face}
    n = face.normal
    points, vecs = [], []
    # get all points
    ents.length.times{|i| points.push(ents[i]) if ents[i].class == Geom::Point3d
      points.concat(ents[i].vertices.collect{|v| v.position}) if ents[i].respond_to?("vertices")
    }
    points.uniq!
    y_vec = Geom::Vector3d.new( ((n!=[0,0,1])?[0,0,1]:[0,1,0]).project_to_plane([0,0,0], n) ).normalize
    x_vec = y_vec * n
    @t = Geom::Transformation.new(x_vec, y_vec, n, points[0])
    ti = @t.inverse
    points.each{|p| p.transform!(ti)}
    b = Geom::BoundingBox.new
    b.add(*points)
    @b_max = b.max
    @b_min = b.min
  end

  def width
    return (@b_max.x - @b_min.x).abs
  end

  def height
    return (@b_max.z - @b_min.z).abs
  end

  def corner(number)
    p = number.to_s(2).rjust(3).split(//).collect{|s| s.to_i}.reverse
    x = @b_min.x + p.x * (@b_max.x - @b_min.x)
    y = @b_min.y + p.y * (@b_max.y - @b_min.y)
    z = @b_min.z + p.z * (@b_max.z - @b_min.z)
    return Geom::Point3d.new(x,y,z).transform!(@t)
  end


end #class AE::Bounds2



end #module AE



file_loaded(File.basename(__FILE__))
